import json
from typing import List, Dict, Optional

from pydantic.main import BaseModel
from socketio import AsyncServer

from server.game.manager import ClientManager
from server.game.message import MessageToPlayer
from server.model.game import Game
from server.model.players import Player, Announce


class Metadata(BaseModel):
    ready: bool = False
    sid: str = ""
    last_announce: Optional[Announce] = None


class LobbyManager(ClientManager):
    """ A ClientManager that handles a lobby, then orchestrates games.
    """

    def __init__(self, sio: AsyncServer):
        super().__init__()
        self.lobby: Dict[str, Player] = {}
        self.metadata: Dict[str, Metadata] = {}
        self.games: List[Game] = []
        self.sio = sio

    @property
    def nb_ready(self):
        return len(self.players_ready)

    @property
    def players_ready(self):
        return [self.lobby[k] for k, m in self.metadata.items() if m.ready]

    async def add_player(self, player: Player, is_ready: bool = False) -> str:
        self.lobby[player.name] = player
        self.metadata[player.name] = Metadata(ready=is_ready)
        print(f"Added {player} to a lobby with {len(self.lobby)} players.")
        return f"Bienvenu, {player}! Il y a {len(self.lobby)} joueurs en ligne."

    def wants_to_play(self, player: Player):
        self.metadata[player.name].ready = True
        print(f"{player} ready to play! ({self.nb_ready} people ready)")

    def announces(self, player: Player, announce: Announce):
        # FIXME: Call this message on incoming "ANNOUNCE"
        self.metadata[player.name].last_announce = announce
        print(f"{player} ready to play! ({self.nb_ready} people ready)")

    def start_game(self, max_players=2):
        players = self.players_ready[:max_players]
        self.games.append(Game(players, manager=self))
        for p in players:
            self.metadata[p.name].ready = False

        print(f"Game started : {' vs '.join([p.name for p in players])}")

    def send(self, to: Player, message: MessageToPlayer, extra=None):
        self.sio.send(message)
        pass

    def handle_message(self, sid, data):
        sender: Optional[Player] = None
        sanitized = str(data)
        print(f"Lobby| Received message from {sid}: {data}.")

        for player in self.players:
            if player.name in sanitized:
                sender = player
        if sender:
            print(f"Lobby| Found sender: {sender.name}")

        message = MessageToPlayer.Waiting
        extras = [p.name for p in self.players]
        body = {"message": message.name}
        if extras:
            body["extras"] = extras
        return json.dumps(body)
